# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from fastapi import HTTPException, status
from sqlalchemy import asc, desc

if TYPE_CHECKING:
    from sqlalchemy.orm import InstrumentedAttribute
    from sqlalchemy.sql.elements import ColumnElement


def build_ordering(
    order_by: str, *, allowed: Mapping[str, ColumnElement[Any]] | Mapping[str, InstrumentedAttribute[Any]]
) -> ColumnElement[Any]:
    """
    Build an SQLAlchemy ORDER BY expression from the `order_by` parameter.

    :param order_by: Public field name, optionally prefixed with "-" for descending.
    :param allowed: Map of public field to SQLAlchemy column/expression.
    """
    is_desc = order_by.startswith("-")
    key = order_by.lstrip("-")

    if key not in allowed:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Ordering with '{order_by}' is disallowed or the attribute does not exist on the model",
        )

    col = allowed[key]
    return desc(col) if is_desc else asc(col)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest.mock import ANY, MagicMock, patch

import pytest
from starlette.exceptions import HTTPException

from airflow.providers.fab.auth_manager.api_fastapi.services.login import FABAuthManagerLogin


@pytest.fixture
def auth_manager():
    return MagicMock()


@pytest.fixture
def user():
    user = MagicMock()
    user.password = "dummy"
    return user


@patch("airflow.providers.fab.auth_manager.api_fastapi.services.login.get_fab_auth_manager")
class TestLogin:
    def setup_method(
        self,
    ):
        self.login_body = {"username": "username", "password": "password"}
        self.dummy_token = "DUMMY_TOKEN"

    def test_create_token(self, get_auth_manager, auth_manager, user):
        auth_manager.create_token.return_value = user
        auth_manager.generate_jwt.return_value = self.dummy_token

        get_auth_manager.return_value = auth_manager

        result = FABAuthManagerLogin.create_token(
            headers={},
            body=self.login_body,
        )
        assert result.access_token == self.dummy_token
        auth_manager.create_token.assert_called_once_with(headers={}, body=self.login_body)
        auth_manager.generate_jwt.assert_called_once_with(user=user, expiration_time_in_seconds=ANY)

    def test_create_token_no_user(self, get_auth_manager, auth_manager):
        auth_manager.create_token.return_value = None
        get_auth_manager.return_value = auth_manager

        with pytest.raises(HTTPException, match="Invalid credentials") as ex:
            FABAuthManagerLogin.create_token(
                headers={},
                body=self.login_body,
            )
        assert ex.value.status_code == 401

    def test_create_token_wrong_payload(self, get_auth_manager, auth_manager):
        auth_manager.create_token.side_effect = ValueError("test")
        get_auth_manager.return_value = auth_manager

        with pytest.raises(HTTPException, match="test") as ex:
            FABAuthManagerLogin.create_token(
                headers={},
                body=self.login_body,
            )
        assert ex.value.status_code == 400



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import functools
from unittest.mock import patch

from airflow.providers.fab.www.app import purge_cached_app


def dont_initialize_flask_app_submodules(_func=None, *, skip_all_except=None):
    if not skip_all_except:
        skip_all_except = []

    def decorator_dont_initialize_flask_app_submodules(f):
        def no_op(*args, **kwargs):
            pass

        methods = [
            "init_api_auth",
            "init_plugins",
            "init_error_handlers",
            "init_api_auth_provider",
            "init_api_error_handlers",
            "init_jinja_globals",
            "init_airflow_session_interface",
            "init_appbuilder",
        ]

        @functools.wraps(f)
        def func(*args, **kwargs):
            for method in methods:
                if method not in skip_all_except:
                    patcher = patch(f"airflow.providers.fab.www.app.{method}", no_op)
                    patcher.start()
            purge_cached_app()
            result = f(*args, **kwargs)
            patch.stopall()
            purge_cached_app()

            return result

        return func

    if _func is None:
        return decorator_dont_initialize_flask_app_submodules
    return decorator_dont_initialize_flask_app_submodules(_func)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import unittest.mock

import pytest
from flask_login import logout_user
from sqlalchemy import delete, func, select

from airflow.providers.fab.www.api_connexion.exceptions import EXCEPTIONS_LINK_MAP
from airflow.providers.fab.www.security import permissions
from airflow.utils.session import create_session

from tests_common.test_utils.config import conf_vars
from unit.fab.auth_manager.api_endpoints.api_connexion_utils import (
    assert_401,
    create_user,
    delete_role,
    delete_user,
)

try:
    from airflow.utils import timezone  # type: ignore[attr-defined]
except AttributeError:
    from airflow.sdk import timezone


from airflow.providers.fab.auth_manager.models import User

pytestmark = pytest.mark.db_test


DEFAULT_TIME = "2020-06-11T18:00:00+00:00"


@pytest.fixture(scope="module")
def configured_app(minimal_app_for_auth_api):
    with conf_vars(
        {
            (
                "core",
                "auth_manager",
            ): "airflow.providers.fab.auth_manager.fab_auth_manager.FabAuthManager",
        }
    ):
        with minimal_app_for_auth_api.app_context():
            create_user(
                minimal_app_for_auth_api,
                username="test",
                role_name="Test",
                permissions=[
                    (permissions.ACTION_CAN_CREATE, permissions.RESOURCE_USER),
                    (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_USER),
                    (permissions.ACTION_CAN_EDIT, permissions.RESOURCE_USER),
                    (permissions.ACTION_CAN_READ, permissions.RESOURCE_USER),
                ],
            )
            create_user(
                minimal_app_for_auth_api, username="test_no_permissions", role_name="TestNoPermissions"
            )

            yield minimal_app_for_auth_api

            delete_user(minimal_app_for_auth_api, username="test")
            delete_user(minimal_app_for_auth_api, username="test_no_permissions")
            delete_role(minimal_app_for_auth_api, name="TestNoPermissions")


class TestUserEndpoint:
    @pytest.fixture(autouse=True)
    def setup_attrs(self, configured_app, request) -> None:
        self.app = configured_app
        self.client = self.app.test_client()
        self.session = self.app.appbuilder.session

        # Logout the user after each request
        @request.addfinalizer
        def logout():
            with configured_app.test_request_context():
                logout_user()

    def teardown_method(self) -> None:
        # Delete users that have our custom default time
        self.session.execute(delete(User).where(User.changed_on == timezone.parse(DEFAULT_TIME)))
        self.session.commit()

    def _create_users(self, count, roles=None):
        # create users with defined created_on and changed_on date
        # for easy testing
        if roles is None:
            roles = []
        return [
            User(
                first_name=f"test{i}",
                last_name=f"test{i}",
                username=f"TEST_USER{i}",
                email=f"mytest@test{i}.org",
                roles=roles or [],
                created_on=timezone.parse(DEFAULT_TIME),
                changed_on=timezone.parse(DEFAULT_TIME),
                active=True,
            )
            for i in range(1, count + 1)
        ]


class TestGetUser(TestUserEndpoint):
    def test_should_respond_200(self):
        users = self._create_users(1)
        self.session.add_all(users)
        self.session.commit()
        response = self.client.get("/fab/v1/users/TEST_USER1", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert response.json == {
            "active": True,
            "changed_on": DEFAULT_TIME,
            "created_on": DEFAULT_TIME,
            "email": "mytest@test1.org",
            "fail_login_count": None,
            "first_name": "test1",
            "last_login": None,
            "last_name": "test1",
            "login_count": None,
            "roles": [],
            "username": "TEST_USER1",
        }

    def test_last_names_can_be_empty(self):
        prince = User(
            first_name="Prince",
            last_name="",
            username="prince",
            email="prince@example.org",
            roles=[],
            created_on=timezone.parse(DEFAULT_TIME),
            changed_on=timezone.parse(DEFAULT_TIME),
        )
        self.session.add_all([prince])
        self.session.commit()
        response = self.client.get("/fab/v1/users/prince", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert response.json == {
            "active": True,
            "changed_on": DEFAULT_TIME,
            "created_on": DEFAULT_TIME,
            "email": "prince@example.org",
            "fail_login_count": None,
            "first_name": "Prince",
            "last_login": None,
            "last_name": "",
            "login_count": None,
            "roles": [],
            "username": "prince",
        }

    def test_first_names_can_be_empty(self):
        liberace = User(
            first_name="",
            last_name="Liberace",
            username="liberace",
            email="liberace@example.org",
            roles=[],
            created_on=timezone.parse(DEFAULT_TIME),
            changed_on=timezone.parse(DEFAULT_TIME),
        )
        self.session.add_all([liberace])
        self.session.commit()
        response = self.client.get("/fab/v1/users/liberace", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert response.json == {
            "active": True,
            "changed_on": DEFAULT_TIME,
            "created_on": DEFAULT_TIME,
            "email": "liberace@example.org",
            "fail_login_count": None,
            "first_name": "",
            "last_login": None,
            "last_name": "Liberace",
            "login_count": None,
            "roles": [],
            "username": "liberace",
        }

    def test_both_first_and_last_names_can_be_empty(self):
        nameless = User(
            first_name="",
            last_name="",
            username="nameless",
            email="nameless@example.org",
            roles=[],
            created_on=timezone.parse(DEFAULT_TIME),
            changed_on=timezone.parse(DEFAULT_TIME),
        )
        self.session.add_all([nameless])
        self.session.commit()
        response = self.client.get("/fab/v1/users/nameless", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert response.json == {
            "active": True,
            "changed_on": DEFAULT_TIME,
            "created_on": DEFAULT_TIME,
            "email": "nameless@example.org",
            "fail_login_count": None,
            "first_name": "",
            "last_login": None,
            "last_name": "",
            "login_count": None,
            "roles": [],
            "username": "nameless",
        }

    def test_should_respond_404(self):
        response = self.client.get("/fab/v1/users/invalid-user", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 404
        assert response.json == {
            "detail": "The User with username `invalid-user` was not found",
            "status": 404,
            "title": "User not found",
            "type": EXCEPTIONS_LINK_MAP[404],
        }

    def test_should_raises_401_unauthenticated(self):
        response = self.client.get("/fab/v1/users/TEST_USER1")
        assert_401(response)

    def test_should_raise_403_forbidden(self):
        response = self.client.get(
            "/fab/v1/users/TEST_USER1", environ_overrides={"REMOTE_USER": "test_no_permissions"}
        )
        assert response.status_code == 403


class TestGetUsers(TestUserEndpoint):
    def test_should_response_200(self):
        response = self.client.get("/fab/v1/users", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert response.json["total_entries"] == 2
        usernames = [user["username"] for user in response.json["users"] if user]
        assert usernames == ["test", "test_no_permissions"]

    def test_should_raises_401_unauthenticated(self):
        response = self.client.get("/fab/v1/users")
        assert_401(response)

    def test_should_raise_403_forbidden(self):
        response = self.client.get("/fab/v1/users", environ_overrides={"REMOTE_USER": "test_no_permissions"})
        assert response.status_code == 403


class TestGetUsersPagination(TestUserEndpoint):
    @pytest.mark.parametrize(
        ("url", "expected_usernames"),
        [
            ("/fab/v1/users?limit=1", ["test"]),
            ("/fab/v1/users?limit=2", ["test", "test_no_permissions"]),
            (
                "/fab/v1/users?offset=5",
                [
                    "TEST_USER4",
                    "TEST_USER5",
                    "TEST_USER6",
                    "TEST_USER7",
                    "TEST_USER8",
                    "TEST_USER9",
                    "TEST_USER10",
                ],
            ),
            (
                "/fab/v1/users?offset=0",
                [
                    "test",
                    "test_no_permissions",
                    "TEST_USER1",
                    "TEST_USER2",
                    "TEST_USER3",
                    "TEST_USER4",
                    "TEST_USER5",
                    "TEST_USER6",
                    "TEST_USER7",
                    "TEST_USER8",
                    "TEST_USER9",
                    "TEST_USER10",
                ],
            ),
            ("/fab/v1/users?limit=1&offset=5", ["TEST_USER4"]),
            ("/fab/v1/users?limit=1&offset=1", ["test_no_permissions"]),
            (
                "/fab/v1/users?limit=2&offset=2",
                ["TEST_USER1", "TEST_USER2"],
            ),
        ],
    )
    def test_handle_limit_offset(self, url, expected_usernames):
        users = self._create_users(10)
        self.session.add_all(users)
        self.session.commit()
        response = self.client.get(url, environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert response.json["total_entries"] == 12
        usernames = [user["username"] for user in response.json["users"] if user]
        assert usernames == expected_usernames

    def test_should_respect_page_size_limit_default(self):
        users = self._create_users(200)
        self.session.add_all(users)
        self.session.commit()

        response = self.client.get("/fab/v1/users", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        # Explicitly add the 2 users on setUp
        assert response.json["total_entries"] == 200 + len(["test", "test_no_permissions"])
        assert len(response.json["users"]) == 100

    def test_should_response_400_with_invalid_order_by(self):
        users = self._create_users(2)
        self.session.add_all(users)
        self.session.commit()
        response = self.client.get("/fab/v1/users?order_by=myname", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 400
        msg = "Ordering with 'myname' is disallowed or the attribute does not exist on the model"
        assert response.json["detail"] == msg

    def test_limit_of_zero_should_return_default(self):
        users = self._create_users(200)
        self.session.add_all(users)
        self.session.commit()

        response = self.client.get("/fab/v1/users?limit=0", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        # Explicit add the 2 users on setUp
        assert response.json["total_entries"] == 200 + len(["test", "test_no_permissions"])
        assert len(response.json["users"]) == 50

    @conf_vars({("api", "maximum_page_limit"): "150"})
    def test_should_return_conf_max_if_req_max_above_conf(self):
        users = self._create_users(200)
        self.session.add_all(users)
        self.session.commit()

        response = self.client.get("/fab/v1/users?limit=180", environ_overrides={"REMOTE_USER": "test"})
        assert response.status_code == 200
        assert len(response.json["users"]) == 150


EXAMPLE_USER_NAME = "example_user"

EXAMPLE_USER_EMAIL = "example_user@example.com"


def _delete_user(**filters):
    with create_session() as session:
        user = session.scalars(select(User).filter_by(**filters)).first()
        if user is None:
            return
        user.roles = []
        session.execute(delete(User).filter_by(**filters))


@pytest.fixture
def autoclean_username():
    _delete_user(username=EXAMPLE_USER_NAME)
    yield EXAMPLE_USER_NAME
    _delete_user(username=EXAMPLE_USER_NAME)


@pytest.fixture
def autoclean_email():
    _delete_user(email=EXAMPLE_USER_EMAIL)
    yield EXAMPLE_USER_EMAIL
    _delete_user(email=EXAMPLE_USER_EMAIL)


@pytest.fixture
def user_with_same_username(configured_app, autoclean_username):
    user = create_user(
        configured_app,
        username=autoclean_username,
        email="another_user@example.com",
        role_name="TestNoPermissions",
    )
    assert user, f"failed to create user '{autoclean_username} <another_user@example.com>'"
    return user


@pytest.fixture
def user_with_same_email(configured_app, autoclean_email):
    user = create_user(
        configured_app,
        username="another_user",
        email=autoclean_email,
        role_name="TestNoPermissions",
    )
    assert user, f"failed to create user 'another_user <{autoclean_email}>'"
    return user


@pytest.fixture
def user_different(configured_app):
    username = "another_user"
    email = "another_user@example.com"

    _delete_user(username=username, email=email)
    user = create_user(configured_app, username=username, email=email, role_name="TestNoPermissions")
    assert user, "failed to create user 'another_user <another_user@example.com>'"
    yield user
    _delete_user(username=username, email=email)


@pytest.fixture
def autoclean_user_payload(autoclean_username, autoclean_email):
    return {
        "username": autoclean_username,
        "password": "resutsop",
        "email": autoclean_email,
        "first_name": "Tester",
        "last_name": "",
    }


@pytest.fixture
def autoclean_admin_user(configured_app, autoclean_user_payload):
    security_manager = configured_app.appbuilder.sm
    return security_manager.add_user(
        role=security_manager.find_role("Admin"),
        **autoclean_user_payload,
    )


class TestPostUser(TestUserEndpoint):
    def test_with_default_role(self, autoclean_username, autoclean_user_payload):
        self.client.application.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        response = self.client.post(
            "/fab/v1/users",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json

        security_manager = self.app.appbuilder.sm
        user = security_manager.find_user(autoclean_username)
        assert user is not None
        assert user.roles == [security_manager.find_role("Public")]

    def test_with_custom_roles(self, autoclean_username, autoclean_user_payload):
        response = self.client.post(
            "/fab/v1/users",
            json={"roles": [{"name": "User"}, {"name": "Viewer"}], **autoclean_user_payload},
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json

        security_manager = self.app.appbuilder.sm
        user = security_manager.find_user(autoclean_username)
        assert user is not None
        assert {r.name for r in user.roles} == {"User", "Viewer"}

    @pytest.mark.usefixtures("user_different")
    def test_with_existing_different_user(self, autoclean_user_payload):
        response = self.client.post(
            "/fab/v1/users",
            json={"roles": [{"name": "User"}, {"name": "Viewer"}], **autoclean_user_payload},
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json

    def test_unauthenticated(self, autoclean_user_payload):
        response = self.client.post(
            "/fab/v1/users",
            json=autoclean_user_payload,
        )
        assert response.status_code == 401, response.json

    def test_forbidden(self, autoclean_user_payload):
        response = self.client.post(
            "/fab/v1/users",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test_no_permissions"},
        )
        assert response.status_code == 403, response.json

    @pytest.mark.parametrize(
        ("existing_user_fixture_name", "error_detail_template"),
        [
            ("user_with_same_username", "Username `{username}` already exists. Use PATCH to update."),
            ("user_with_same_email", "The email `{email}` is already taken."),
        ],
        ids=["username", "email"],
    )
    def test_already_exists(
        self,
        request,
        autoclean_user_payload,
        existing_user_fixture_name,
        error_detail_template,
    ):
        existing = request.getfixturevalue(existing_user_fixture_name)

        response = self.client.post(
            "/fab/v1/users",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 409, response.json

        error_detail = error_detail_template.format(username=existing.username, email=existing.email)
        assert response.json["detail"] == error_detail

    @pytest.mark.parametrize(
        ("payload_converter", "error_message"),
        [
            pytest.param(
                lambda p: {k: v for k, v in p.items() if k != "username"},
                "{'username': ['Missing data for required field.']}",
                id="missing-required",
            ),
            pytest.param(
                lambda p: {"i-am": "a typo", **p},
                "{'i-am': ['Unknown field.']}",
                id="unknown-user-field",
            ),
            pytest.param(
                lambda p: {**p, "roles": [{"also": "a typo", "name": "User"}]},
                "{'roles': {0: {'also': ['Unknown field.']}}}",
                id="unknown-role-field",
            ),
            pytest.param(
                lambda p: {**p, "roles": [{"name": "God"}, {"name": "User"}, {"name": "Overlord"}]},
                "Unknown roles: 'God', 'Overlord'",
                id="unknown-role",
            ),
        ],
    )
    def test_invalid_payload(self, autoclean_user_payload, payload_converter, error_message):
        response = self.client.post(
            "/fab/v1/users",
            json=payload_converter(autoclean_user_payload),
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 400, response.json
        assert response.json == {
            "detail": error_message,
            "status": 400,
            "title": "Bad Request",
            "type": EXCEPTIONS_LINK_MAP[400],
        }

    def test_internal_server_error(self, autoclean_user_payload):
        with unittest.mock.patch.object(self.app.appbuilder.sm, "add_user", return_value=None):
            response = self.client.post(
                "/fab/v1/users",
                json=autoclean_user_payload,
                environ_overrides={"REMOTE_USER": "test"},
            )
            assert response.json == {
                "detail": "Failed to add user `example_user`.",
                "status": 500,
                "title": "Internal Server Error",
                "type": EXCEPTIONS_LINK_MAP[500],
            }


class TestPatchUser(TestUserEndpoint):
    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_change(self, autoclean_username, autoclean_user_payload):
        autoclean_user_payload["first_name"] = "Changed"
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json

        # The first name is changed.
        data = response.json
        assert data["first_name"] == "Changed"
        assert data["last_name"] == ""

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_change_with_update_mask(self, autoclean_username, autoclean_user_payload):
        autoclean_user_payload["first_name"] = "Changed"
        autoclean_user_payload["last_name"] = "McTesterson"
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}?update_mask=last_name",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json

        # The first name is changed, but the last name isn't since we masked it.
        data = response.json
        assert data["first_name"] == "Tester"
        assert data["last_name"] == "McTesterson"

    @pytest.mark.parametrize(
        ("payload", "error_message"),
        [
            ({"username": "another_user"}, "The username `another_user` already exists"),
            ({"email": "another_user@example.com"}, "The email `another_user@example.com` already exists"),
        ],
        ids=["username", "email"],
    )
    @pytest.mark.usefixtures("user_different")
    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_patch_already_exists(
        self,
        payload,
        error_message,
        autoclean_user_payload,
        autoclean_username,
    ):
        autoclean_user_payload.update(payload)
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 409, response.json

        assert response.json["detail"] == error_message

    @pytest.mark.parametrize(
        "field",
        ["username", "first_name", "last_name", "email"],
    )
    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_required_fields(
        self,
        field,
        autoclean_user_payload,
        autoclean_username,
    ):
        autoclean_user_payload.pop(field)
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 400, response.json
        assert response.json["detail"] == f"{{'{field}': ['Missing data for required field.']}}"

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_username_can_be_updated(self, autoclean_user_payload, autoclean_username):
        testusername = "testusername"
        autoclean_user_payload.update({"username": testusername})
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        _delete_user(username=testusername)
        assert response.json["username"] == testusername

    @pytest.mark.usefixtures("autoclean_admin_user")
    @unittest.mock.patch(
        "airflow.providers.fab.auth_manager.api_endpoints.user_endpoint.generate_password_hash",
        return_value="fake-hashed-pass",
    )
    def test_password_hashed(
        self,
        mock_generate_password_hash,
        autoclean_username,
        autoclean_user_payload,
    ):
        autoclean_user_payload["password"] = "new-pass"
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json
        assert "password" not in response.json

        mock_generate_password_hash.assert_called_once_with("new-pass")
        password_in_db = self.session.scalar(select(User.password).where(User.username == autoclean_username))
        assert password_in_db == "fake-hashed-pass"

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_replace_roles(self, autoclean_username, autoclean_user_payload):
        # Patching a user's roles should replace the entire list.
        autoclean_user_payload["roles"] = [{"name": "User"}, {"name": "Viewer"}]
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}?update_mask=roles",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json
        assert {d["name"] for d in response.json["roles"]} == {"User", "Viewer"}

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_unchanged(self, autoclean_username, autoclean_user_payload):
        # Should allow a PATCH that changes nothing.
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 200, response.json

        expected = {k: v for k, v in autoclean_user_payload.items() if k != "password"}
        assert {k: response.json[k] for k in expected} == expected

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_unauthenticated(self, autoclean_username, autoclean_user_payload):
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
        )
        assert response.status_code == 401, response.json

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_forbidden(self, autoclean_username, autoclean_user_payload):
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test_no_permissions"},
        )
        assert response.status_code == 403, response.json

    def test_not_found(self, autoclean_username, autoclean_user_payload):
        # This test does not populate autoclean_admin_user into the database.
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=autoclean_user_payload,
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 404, response.json

    @pytest.mark.parametrize(
        ("payload_converter", "error_message"),
        [
            pytest.param(
                lambda p: {k: v for k, v in p.items() if k != "username"},
                "{'username': ['Missing data for required field.']}",
                id="missing-required",
            ),
            pytest.param(
                lambda p: {"i-am": "a typo", **p},
                "{'i-am': ['Unknown field.']}",
                id="unknown-user-field",
            ),
            pytest.param(
                lambda p: {**p, "roles": [{"also": "a typo", "name": "User"}]},
                "{'roles': {0: {'also': ['Unknown field.']}}}",
                id="unknown-role-field",
            ),
            pytest.param(
                lambda p: {**p, "roles": [{"name": "God"}, {"name": "User"}, {"name": "Overlord"}]},
                "Unknown roles: 'God', 'Overlord'",
                id="unknown-role",
            ),
        ],
    )
    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_invalid_payload(
        self,
        autoclean_username,
        autoclean_user_payload,
        payload_converter,
        error_message,
    ):
        response = self.client.patch(
            f"/fab/v1/users/{autoclean_username}",
            json=payload_converter(autoclean_user_payload),
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 400, response.json
        assert response.json == {
            "detail": error_message,
            "status": 400,
            "title": "Bad Request",
            "type": EXCEPTIONS_LINK_MAP[400],
        }


class TestDeleteUser(TestUserEndpoint):
    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_delete(self, autoclean_username):
        response = self.client.delete(
            f"/fab/v1/users/{autoclean_username}",
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 204, response.json  # NO CONTENT.
        assert (
            self.session.scalar(select(func.count(User.id)).where(User.username == autoclean_username)) == 0
        )

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_unauthenticated(self, autoclean_username):
        response = self.client.delete(
            f"/fab/v1/users/{autoclean_username}",
        )
        assert response.status_code == 401, response.json
        assert (
            self.session.scalar(select(func.count(User.id)).where(User.username == autoclean_username)) == 1
        )

    @pytest.mark.usefixtures("autoclean_admin_user")
    def test_forbidden(self, autoclean_username):
        response = self.client.delete(
            f"/fab/v1/users/{autoclean_username}",
            environ_overrides={"REMOTE_USER": "test_no_permissions"},
        )
        assert response.status_code == 403, response.json
        assert (
            self.session.scalar(select(func.count(User.id)).where(User.username == autoclean_username)) == 1
        )

    def test_not_found(self, autoclean_username):
        # This test does not populate autoclean_admin_user into the database.
        response = self.client.delete(
            f"/fab/v1/users/{autoclean_username}",
            environ_overrides={"REMOTE_USER": "test"},
        )
        assert response.status_code == 404, response.json



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from airflow.providers.fab.www.utils import get_session_lifetime_config

from tests_common.test_utils.config import conf_vars


class TestUpdatedConfigNames:
    @conf_vars({("fab", "session_lifetime_minutes"): "43200"})
    def test_config_val_is_default(self):
        session_lifetime_config = get_session_lifetime_config()
        assert session_lifetime_config == 43200

    @conf_vars({("fab", "session_lifetime_minutes"): "43201"})
    def test_config_val_is_not_default(self):
        session_lifetime_config = get_session_lifetime_config()
        assert session_lifetime_config == 43201

    @conf_vars({("fab", "session_lifetime_days"): ""})
    def test_uses_updated_session_timeout_config_by_default(self):
        session_lifetime_config = get_session_lifetime_config()
        default_timeout_minutes = 30 * 24 * 60
        assert session_lifetime_config == default_timeout_minutes



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest.mock import patch

import pytest

from airflow.api_fastapi.auth.managers.base_auth_manager import COOKIE_NAME_JWT_TOKEN
from airflow.providers.fab.auth_manager.api_fastapi.datamodels.login import LoginResponse


@pytest.mark.db_test
class TestLogin:
    dummy_login_body = {"username": "dummy", "password": "dummy"}
    dummy_token = LoginResponse(access_token="DUMMY_TOKEN")

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.login.FABAuthManagerLogin")
    def test_create_token(self, mock_fab_auth_manager_login, test_client):
        mock_fab_auth_manager_login.create_token.return_value = self.dummy_token

        response = test_client.post(
            "/token",
            json=self.dummy_login_body,
        )
        assert response.status_code == 201
        assert response.json()["access_token"] == self.dummy_token.access_token

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.login.FABAuthManagerLogin")
    def test_create_token_cli(self, mock_fab_auth_manager_login, test_client):
        mock_fab_auth_manager_login.create_token.return_value = LoginResponse(access_token="DUMMY_TOKEN")

        response = test_client.post(
            "/token/cli",
            json=self.dummy_login_body,
        )
        assert response.status_code == 201
        assert response.json()["access_token"] == self.dummy_token.access_token

    def test_logout(self, test_client):
        response = test_client.get("/logout", follow_redirects=False)
        assert response.status_code == 307
        assert response.headers["location"] == "/auth/login"
        cookies = response.headers.get_list("set-cookie")
        assert any("session=" in c for c in cookies)
        assert any(f"{COOKIE_NAME_JWT_TOKEN}=" in c for c in cookies)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

import airflow.providers.fab.www.auth as auth
from airflow.api_fastapi.auth.managers.models.resource_details import DagAccessEntity
from airflow.models import Connection, Pool, Variable
from airflow.providers.fab.www import app as application

from tests_common.test_utils.config import conf_vars

mock_call = Mock()


@pytest.fixture
def app():
    with conf_vars(
        {
            (
                "core",
                "auth_manager",
            ): "airflow.providers.fab.auth_manager.fab_auth_manager.FabAuthManager",
        }
    ):
        yield application.create_app(enable_plugins=False)


@pytest.mark.parametrize(
    ("decorator_name", "is_authorized_method_name"),
    [
        ("has_access_configuration", "is_authorized_configuration"),
        ("has_access_asset", "is_authorized_asset"),
        ("has_access_view", "is_authorized_view"),
    ],
)
class TestHasAccessNoDetails:
    def setup_method(self):
        mock_call.reset_mock()

    def method_test(self):
        mock_call()
        return True

    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    @patch("airflow.providers.fab.www.auth.get_fab_auth_manager")
    def test_has_access_no_details_when_authorized(
        self, _, mock_get_auth_manager, decorator_name, is_authorized_method_name
    ):
        auth_manager = Mock()
        is_authorized_method = Mock()
        is_authorized_method.return_value = True
        setattr(auth_manager, is_authorized_method_name, is_authorized_method)
        mock_get_auth_manager.return_value = auth_manager

        result = getattr(auth, decorator_name)("GET")(self.method_test)()

        mock_call.assert_called_once()
        assert result is True

    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    @patch("airflow.providers.fab.www.auth.get_fab_auth_manager")
    @patch("airflow.providers.fab.www.auth.render_template")
    def test_has_access_no_details_when_no_permission(
        self, mock_render_template, _, mock_get_auth_manager, decorator_name, is_authorized_method_name
    ):
        auth_manager = Mock()
        is_authorized_method = Mock()
        is_authorized_method.return_value = False
        setattr(auth_manager, is_authorized_method_name, is_authorized_method)
        auth_manager.is_logged_in.return_value = True
        auth_manager.is_authorized_view.return_value = False
        mock_get_auth_manager.return_value = auth_manager

        getattr(auth, decorator_name)("GET")(self.method_test)()

        mock_call.assert_not_called()
        mock_render_template.assert_called_once()

    @pytest.mark.db_test
    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    def test_has_access_no_details_when_not_logged_in(
        self, mock_get_auth_manager, app, decorator_name, is_authorized_method_name
    ):
        auth_manager = Mock()
        is_authorized_method = Mock()
        is_authorized_method.return_value = False
        setattr(auth_manager, is_authorized_method_name, is_authorized_method)
        auth_manager.is_logged_in.return_value = False
        auth_manager.get_url_login.return_value = "login_url"
        mock_get_auth_manager.return_value = auth_manager

        with app.test_request_context():
            result = getattr(auth, decorator_name)("GET")(self.method_test)()

        mock_call.assert_not_called()
        assert result.status_code == 302


@pytest.fixture
def get_connection():
    return [Connection("conn_1"), Connection("conn_2")]


@pytest.fixture
def get_pool():
    return [Pool(pool="pool_1"), Pool(pool="pool_2")]


@pytest.fixture
def get_variable():
    return [Variable("var_1"), Variable("var_2")]


@pytest.mark.parametrize(
    ("decorator_name", "is_authorized_method_name", "items"),
    [
        (
            "has_access_connection",
            "is_authorized_connection",
            "get_connection",
        ),
        ("has_access_pool", "is_authorized_pool", "get_pool"),
        ("has_access_variable", "is_authorized_variable", "get_variable"),
    ],
)
class TestHasAccessWithDetails:
    def setup_method(self):
        mock_call.reset_mock()

    def method_test(self, _view, arg):
        mock_call()
        return True

    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    def test_has_access_with_details_when_authorized(
        self, mock_get_auth_manager, decorator_name, is_authorized_method_name, items, request
    ):
        items = request.getfixturevalue(items)
        auth_manager = Mock()
        is_authorized_method = Mock()
        is_authorized_method.return_value = True
        setattr(auth_manager, is_authorized_method_name, is_authorized_method)
        mock_get_auth_manager.return_value = auth_manager

        result = getattr(auth, decorator_name)("GET")(self.method_test)(None, items)

        mock_call.assert_called_once()
        assert result is True

    @pytest.mark.db_test
    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    def test_has_access_with_details_when_unauthorized(
        self, mock_get_auth_manager, app, decorator_name, is_authorized_method_name, items, request
    ):
        items = request.getfixturevalue(items)
        auth_manager = Mock()
        is_authorized_method = Mock()
        is_authorized_method.return_value = False
        setattr(auth_manager, is_authorized_method_name, is_authorized_method)
        mock_get_auth_manager.return_value = auth_manager

        with app.test_request_context():
            result = getattr(auth, decorator_name)("GET")(self.method_test)(None, items)

        mock_call.assert_not_called()
        assert result.status_code == 302


@pytest.mark.parametrize(
    "dag_access_entity",
    [
        DagAccessEntity.XCOM,
        DagAccessEntity.RUN,
        DagAccessEntity.TASK_INSTANCE,
    ],
)
class TestHasAccessDagEntities:
    def setup_method(self):
        mock_call.reset_mock()

    def method_test(self, _view, arg):
        mock_call()
        return True

    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    def test_has_access_dag_entities_when_authorized(self, mock_get_auth_manager, dag_access_entity):
        auth_manager = Mock()
        auth_manager.batch_is_authorized_dag.return_value = True
        mock_get_auth_manager.return_value = auth_manager
        items = [Mock(dag_id="dag_1"), Mock(dag_id="dag_2")]

        result = auth.has_access_dag_entities("GET", dag_access_entity)(self.method_test)(None, items)

        mock_call.assert_called_once()
        assert result is True

    @pytest.mark.db_test
    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    @patch("airflow.providers.fab.www.auth.get_fab_auth_manager")
    def test_has_access_dag_entities_when_unauthorized(
        self, _, mock_get_auth_manager, app, dag_access_entity
    ):
        auth_manager = Mock()
        auth_manager.batch_is_authorized_dag.return_value = False
        mock_get_auth_manager.return_value = auth_manager
        items = [Mock(dag_id="dag_1"), Mock(dag_id="dag_2")]

        with app.test_request_context():
            result = auth.has_access_dag_entities("GET", dag_access_entity)(self.method_test)(None, items)

        mock_call.assert_not_called()
        assert result.headers["Location"] == "/"

    @pytest.mark.db_test
    @patch("airflow.providers.fab.www.auth.get_auth_manager")
    def test_has_access_dag_entities_when_logged_out(self, mock_get_auth_manager, app, dag_access_entity):
        auth_manager = Mock()
        auth_manager.batch_is_authorized_dag.return_value = False
        auth_manager.is_logged_in.return_value = False
        auth_manager.get_url_login.return_value = "login_url"
        mock_get_auth_manager.return_value = auth_manager
        items = [Mock(dag_id="dag_1"), Mock(dag_id="dag_2")]

        with app.test_request_context():
            result = auth.has_access_dag_entities("GET", dag_access_entity)(self.method_test)(None, items)

        mock_call.assert_not_called()
        assert result.headers["Location"] == "login_url"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging

from fastapi import Query

from airflow.configuration import conf

log = logging.getLogger(__name__)


def get_effective_limit(default: int = 100):
    """
    Return a FastAPI dependency that enforces API page limit rules.

    :param default: Default limit if not provided by client.
    """

    def _limit(
        limit: int = Query(
            default,
            ge=0,
        ),
    ) -> int:
        max_val = conf.getint("api", "maximum_page_limit")
        fallback = conf.getint("api", "fallback_page_limit")

        if limit == 0:
            return fallback
        if limit > max_val:
            log.warning(
                "The limit param value %s passed in API exceeds the configured maximum page limit %s",
                limit,
                max_val,
            )
            return max_val
        return limit

    return _limit



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from pathlib import Path

from airflow.configuration import conf

WWW = Path(__file__).resolve().parent
# There is a difference with configuring Swagger in Connexion 2.x and Connexion 3.x
# Connexion 2: https://connexion.readthedocs.io/en/2.14.2/quickstart.html#the-swagger-ui-console
# Connexion 3: https://connexion.readthedocs.io/en/stable/swagger_ui.html#configuring-the-swagger-ui
SWAGGER_ENABLED = conf.getboolean("api", "enable_swagger_ui", fallback=True)
SWAGGER_BUNDLE = WWW.joinpath("static", "dist", "swagger-ui")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import Mock

import pytest

from airflow.providers.fab.auth_manager.security_manager.override import FabAirflowSecurityManagerOverride


class EmptySecurityManager(FabAirflowSecurityManagerOverride):
    # noinspection PyMissingConstructor
    # super() not called on purpose to avoid the whole chain of init calls
    def __init__(self):
        pass


class TestFabAirflowSecurityManagerOverride:
    def test_load_user(self):
        sm = EmptySecurityManager()
        sm.get_user_by_id = Mock()

        sm.load_user("123")

        sm.get_user_by_id.assert_called_once_with(123)

    @mock.patch("airflow.providers.fab.auth_manager.security_manager.override.g", spec={})
    def test_load_user_jwt(self, mock_g):
        sm = EmptySecurityManager()
        mock_user = Mock()
        sm.load_user = Mock(return_value=mock_user)

        actual_user = sm.load_user_jwt(None, {"sub": "test_identity"})

        sm.load_user.assert_called_once_with("test_identity")
        assert actual_user is mock_user
        assert mock_g.user is mock_user

    @mock.patch("airflow.providers.fab.auth_manager.security_manager.override.check_password_hash")
    def test_check_password(self, check_password):
        sm = EmptySecurityManager()
        mock_user = Mock()
        sm.find_user = Mock(return_value=mock_user)
        check_password.return_value = True
        assert sm.check_password("test_user", "test_password")

    @mock.patch("airflow.providers.fab.auth_manager.security_manager.override.check_password_hash")
    def test_check_password_user_not_found(self, check_password):
        sm = EmptySecurityManager()
        sm.find_user = Mock(return_value=None)
        check_password.return_value = False
        assert not sm.check_password("test_user", "test_password")

    @mock.patch("airflow.providers.fab.auth_manager.security_manager.override.check_password_hash")
    def test_check_password_not_match(self, check_password):
        sm = EmptySecurityManager()
        mock_user = Mock()
        sm.find_user = Mock(return_value=mock_user)
        check_password.return_value = False
        assert not sm.check_password("test_user", "test_password")

    @pytest.mark.parametrize(
        ("provider", "resp", "user_info"),
        [
            ("github", {"login": "test"}, {"username": "github_test"}),
            ("githublocal", {"login": "test"}, {"username": "github_test"}),
            ("twitter", {"screen_name": "test"}, {"username": "twitter_test"}),
            (
                "linkedin",
                {"id": "test", "firstName": "John", "lastName": "Doe", "email-address": "test@example.com"},
                {
                    "username": "linkedin_test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                },
            ),
            (
                "google",
                {"id": "test", "given_name": "John", "family_name": "Doe", "email": "test@example.com"},
                {
                    "username": "google_test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                },
            ),
            (
                "azure",
                {
                    "oid": "test",
                    "given_name": "John",
                    "family_name": "Doe",
                    "email": "test@example.com",
                    "roles": ["admin"],
                },
                {
                    "username": "test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                    "role_keys": ["admin"],
                },
            ),
            (
                "azure",
                {
                    "oid": "test",
                    "given_name": "John",
                    "family_name": "Doe",
                    "upn": "test@example.com",
                    "roles": ["admin"],
                },
                {
                    "username": "test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                    "role_keys": ["admin"],
                },
            ),
            ("openshift", {"metadata": {"name": "test"}}, {"username": "openshift_test"}),
            (
                "okta",
                {
                    "sub": "test",
                    "given_name": "John",
                    "family_name": "Doe",
                    "email": "test@example.com",
                    "groups": ["admin"],
                },
                {
                    "username": "okta_test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                    "role_keys": ["admin"],
                },
            ),
            ("okta", {"error": "access_denied", "error_description": "Invalid bearer token."}, {}),
            (
                "auth0",
                {
                    "sub": "test",
                    "given_name": "John",
                    "family_name": "Doe",
                    "email": "test@example.com",
                    "groups": ["admin"],
                },
                {
                    "username": "auth0_test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                    "role_keys": ["admin"],
                },
            ),
            (
                "keycloak",
                {
                    "preferred_username": "test",
                    "given_name": "John",
                    "family_name": "Doe",
                    "email": "test@example.com",
                    "groups": ["admin"],
                },
                {
                    "username": "test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                    "role_keys": ["admin"],
                },
            ),
            (
                "keycloak_before_17",
                {
                    "preferred_username": "test",
                    "given_name": "John",
                    "family_name": "Doe",
                    "email": "test@example.com",
                    "groups": ["admin"],
                },
                {
                    "username": "test",
                    "first_name": "John",
                    "last_name": "Doe",
                    "email": "test@example.com",
                    "role_keys": ["admin"],
                },
            ),
            (
                "authentik",
                {
                    "nickname": "test",
                    "given_name": "John",
                    "preferred_username": "test@example.com",
                    "groups": ["admin"],
                },
                {
                    "username": "test",
                    "first_name": "John",
                    "email": "test@example.com",
                    "role_keys": ["admin"],
                },
            ),
            (
                "other",
                {"preferred_username": "test", "email": "test@example.com"},
                {
                    "username": "test",
                    "first_name": "",
                    "last_name": "",
                    "email": "test@example.com",
                    "role_keys": [],
                },
            ),
        ],
    )
    def test_get_oauth_user_info(self, provider, resp, user_info):
        sm = EmptySecurityManager()
        sm.appbuilder = Mock(sm=sm)
        sm.oauth_remotes = {}
        sm.oauth_remotes[provider] = Mock(
            get=Mock(return_value=Mock(json=Mock(return_value=resp))),
            userinfo=Mock(return_value=resp),
        )
        sm._decode_and_validate_azure_jwt = Mock(return_value=resp)
        sm._get_authentik_token_info = Mock(return_value=resp)
        assert sm.get_oauth_user_info(provider, {"id_token": None}) == user_info



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import types
from contextlib import contextmanager

import pytest
from fastapi.testclient import TestClient

from airflow.api_fastapi.core_api.security import get_user as get_user_dep
from airflow.providers.fab.auth_manager.fab_auth_manager import FabAuthManager


@pytest.fixture(scope="module")
def fab_auth_manager():
    return FabAuthManager()


@pytest.fixture(scope="module")
def test_client(fab_auth_manager):
    return TestClient(fab_auth_manager.get_fastapi_app())


@pytest.fixture
def override_deps(test_client):
    """
    Context-managed helper for app.dependency_overrides.

    Usage:
        with override_deps({dep_func: override_func, ...}):
            # do requests
    """
    app = test_client.app

    @contextmanager
    def _use(mapping: dict):
        for dep, override in mapping.items():
            app.dependency_overrides[dep] = override
        try:
            yield
        finally:
            for dep in mapping.keys():
                app.dependency_overrides.pop(dep, None)

    return _use


@pytest.fixture
def as_user(override_deps):
    @contextmanager
    def _as(u=types.SimpleNamespace(id=1, username="tester")):
        with override_deps({get_user_dep: lambda: u}):
            yield u

    return _as



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest
from fastapi import HTTPException

from airflow.providers.fab.auth_manager.api_fastapi.security import requires_fab_custom_view


class TestSecurityDependency:
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    def test_requires_fab_custom_view_allows_when_authorized(self, get_auth_manager):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        get_auth_manager.return_value = mgr

        check = requires_fab_custom_view(method="POST", resource_name="Role")
        user = object()

        assert check(user=user) is None
        mgr.is_authorized_custom_view.assert_called_once_with(method="POST", resource_name="Role", user=user)

    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    def test_requires_fab_custom_view_raises_403_when_unauthorized(self, get_auth_manager):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = False
        get_auth_manager.return_value = mgr

        check = requires_fab_custom_view(method="DELETE", resource_name="Role")
        with pytest.raises(HTTPException) as ex:
            check(user=object())
        assert ex.value.status_code == 403



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import HTTPException, status
from sqlalchemy import func, select

from airflow.providers.fab.auth_manager.api_fastapi.datamodels.roles import (
    RoleBody,
    RoleCollectionResponse,
    RoleResponse,
)
from airflow.providers.fab.auth_manager.api_fastapi.sorting import build_ordering
from airflow.providers.fab.auth_manager.models import Role
from airflow.providers.fab.www.utils import get_fab_auth_manager

if TYPE_CHECKING:
    from airflow.providers.fab.auth_manager.security_manager.override import FabAirflowSecurityManagerOverride


class FABAuthManagerRoles:
    """Service layer for FAB Auth Manager role operations (create, validate, sync)."""

    @staticmethod
    def _check_action_and_resource(
        security_manager: FabAirflowSecurityManagerOverride,
        perms: list[tuple[str, str]],
    ) -> None:
        for action_name, resource_name in perms:
            if not security_manager.get_action(action_name):
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=f"The specified action: {action_name!r} was not found",
                )
            if not security_manager.get_resource(resource_name):
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=f"The specified resource: {resource_name!r} was not found",
                )

    @classmethod
    def create_role(cls, body: RoleBody) -> RoleResponse:
        security_manager = get_fab_auth_manager().security_manager

        existing = security_manager.find_role(name=body.name)
        if existing:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Role with name {body.name!r} already exists; please update with the PATCH endpoint",
            )

        perms: list[tuple[str, str]] = [(ar.action.name, ar.resource.name) for ar in (body.permissions or [])]

        cls._check_action_and_resource(security_manager, perms)

        security_manager.bulk_sync_roles([{"role": body.name, "perms": perms}])

        created = security_manager.find_role(name=body.name)
        if not created:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Role was not created due to an unexpected error.",
            )

        return RoleResponse.model_validate(created)

    @classmethod
    def get_roles(cls, *, order_by: str, limit: int, offset: int) -> RoleCollectionResponse:
        security_manager = get_fab_auth_manager().security_manager
        session = security_manager.session

        total_entries = session.scalars(select(func.count(Role.id))).one()

        ordering = build_ordering(order_by, allowed={"name": Role.name, "role_id": Role.id})

        stmt = select(Role).order_by(ordering).offset(offset).limit(limit)
        roles = session.scalars(stmt).unique().all()

        return RoleCollectionResponse(
            roles=[RoleResponse.model_validate(r) for r in roles],
            total_entries=total_entries,
        )

    @classmethod
    def delete_role(cls, name: str) -> None:
        security_manager = get_fab_auth_manager().security_manager

        existing = security_manager.find_role(name=name)
        if not existing:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Role with name {name!r} does not exist.",
            )
        security_manager.delete_role(existing)

    @classmethod
    def get_role(cls, name: str) -> RoleResponse:
        security_manager = get_fab_auth_manager().security_manager

        existing = security_manager.find_role(name=name)
        if not existing:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Role with name {name!r} does not exist.",
            )
        return RoleResponse.model_validate(existing)

    @classmethod
    def patch_role(cls, body: RoleBody, name: str, update_mask: str | None = None) -> RoleResponse:
        security_manager = get_fab_auth_manager().security_manager

        existing = security_manager.find_role(name=name)
        if not existing:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Role with name {name!r} does not exist.",
            )

        if update_mask:
            update_data = RoleResponse.model_validate(existing)

            for field in update_mask:
                if field == "actions":
                    update_data.permissions = body.permissions
                elif hasattr(body, field):
                    setattr(update_data, field, getattr(body, field))
                else:
                    raise HTTPException(
                        status_code=status.HTTP_400_BAD_REQUEST,
                        detail=f"'{field}' in update_mask is unknown",
                    )
        else:
            update_data = RoleResponse(name=body.name, permissions=body.permissions or [])

        perms: list[tuple[str, str]] = [(ar.action.name, ar.resource.name) for ar in (body.permissions or [])]
        cls._check_action_and_resource(security_manager, perms)
        security_manager.bulk_sync_roles([{"role": name, "perms": perms}])

        new_name = update_data.name
        if new_name and new_name != existing.name:
            security_manager.update_role(role_id=existing.id, name=new_name)
        return RoleResponse.model_validate(update_data)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from pydantic import Field

from airflow.api_fastapi.core_api.base import BaseModel, StrictBaseModel


class ActionResponse(BaseModel):
    """Outgoing representation of an action (permission name)."""

    name: str


class ResourceResponse(BaseModel):
    """Outgoing representation of a resource."""

    name: str


class ActionResourceResponse(BaseModel):
    """Pairing of an action with a resource."""

    action: ActionResponse
    resource: ResourceResponse


class RoleBody(StrictBaseModel):
    """Incoming payload for creating/updating a role."""

    name: str = Field(min_length=1)
    permissions: list[ActionResourceResponse] = Field(
        default_factory=list, alias="actions", validation_alias="actions"
    )


class RoleResponse(BaseModel):
    """Outgoing representation of a role and its permissions."""

    name: str
    permissions: list[ActionResourceResponse] = Field(default_factory=list, serialization_alias="actions")


class RoleCollectionResponse(BaseModel):
    """Outgoing representation of a paginated collection of roles."""

    roles: list[RoleResponse]
    total_entries: int



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from flask_appbuilder.security.views import (
    ResetMyPasswordView,
    ResetPasswordView,
    UserInfoEditView,
)

from airflow.providers.fab.www.security import permissions


class CustomUserInfoEditView(UserInfoEditView):
    """Customize permission names for FAB's builtin UserInfoEditView."""

    class_permission_name = permissions.RESOURCE_MY_PROFILE
    route_base = "/userinfoeditview"
    method_permission_name = {
        "this_form_get": "edit",
        "this_form_post": "edit",
    }
    base_permissions = [permissions.ACTION_CAN_EDIT, permissions.ACTION_CAN_READ]


class CustomResetMyPasswordView(ResetMyPasswordView):
    """Customize permission names for FAB's builtin ResetMyPasswordView."""

    class_permission_name = permissions.RESOURCE_MY_PASSWORD
    method_permission_name = {
        "this_form_get": "read",
        "this_form_post": "edit",
    }
    base_permissions = [permissions.ACTION_CAN_EDIT, permissions.ACTION_CAN_READ]


class CustomResetPasswordView(ResetPasswordView):
    """Customize permission names for FAB's builtin ResetPasswordView."""

    class_permission_name = permissions.RESOURCE_PASSWORD
    method_permission_name = {
        "this_form_get": "read",
        "this_form_post": "edit",
    }

    base_permissions = [permissions.ACTION_CAN_EDIT, permissions.ACTION_CAN_READ]
